@@ -1,4 +1,4 @@
-#!/usr/bin/perl 
+#!/usr/bin/perl
 # Script stolen from one of Roderich Schupp's mails to the PAR
 # mailing list. He attributes this to:
 
@@ -21,8 +21,28 @@ the embedded files are extracted into it, using their original file names.
 
 my ($exe, $extract) = @ARGV;
 
-open my $fh, '<', $exe or die qq[failed to open "$exe": $!];
-binmode $fh;
+sub safe_read
+{
+    my ($fh, $n) = @_;
+    my $buf;
+    my $res = read $fh, $buf, $n;
+    die qq[read of $n bytes failed on "$exe": $!] unless defined $res;
+    die qq[read of $n bytes failed on "$exe": at EOF] unless $res > 0;
+    die qq[read of $n bytes failed on "$exe": only read $res bytes] unless $res == $n;
+    return $buf;
+}
+
+sub safe_seek
+{
+    my ($fh, $offset, $whence) = @_;
+    unless (seek $fh, $offset, $whence)
+    {
+        my $what = $whence == 0 ? "SET" : $whence == 1 ? "CUR" : "END";
+        die qq[seek $what of $offset bytes failed on "$exe": $!];
+    }
+}
+
+open my $fh, '<:raw', $exe or die qq[failed to open "$exe": $!];
 
 # search for the "\nPAR.pm\n" signature backward from the end of the file
 my $buf;
@@ -32,34 +52,38 @@ my $idx = -1;
 while (1)
 {
     $offset = $size if $offset > $size;
-    seek $fh, -$offset, 2 or die qq[seek failed on "$exe": $!];
-    my $nread = read $fh, $buf, $offset;
-    die qq[read failed on "$exe": $!] unless $nread == $offset;
-    $idx = rindex($buf, "\nPAR.pm\n");
+    safe_seek($fh, -$offset, 2);
+    $buf = safe_read($fh, $offset);
+    $idx = rindex($buf, "\012PAR.pm\012");
     last if $idx >= 0 || $offset == $size || $offset > 128 * 1024;
     $offset *= 2;
 }
+$offset -= $idx;
 die qq[no PAR signature found in "$exe"] unless $idx >= 0;
 
-# seek 4 bytes backward from the signature to get the offset of the 
+# seek 4 bytes backward from the signature to get the offset of the
 # first embedded FILE, then seek to it
-$offset -= $idx - 4;
-seek $fh, -$offset, 2;
-read $fh, $buf, 4;
-seek $fh, -$offset - unpack("N", $buf), 2;
+$offset += 4;
+safe_seek($fh, -$offset, 2);
+$buf = safe_read($fh, 4);
+safe_seek($fh, -$offset - unpack("N", $buf), 2);
 printf STDERR qq[embedded files in "%s" start at offset %d\n], $exe, tell($fh);
 
-read $fh, $buf, 4;
-while ($buf eq "FILE") 
+my $nfiles = 0;
+$buf = safe_read($fh, 4);
+while ($buf eq "FILE")
 {
-    read $fh, $buf, 4;
-    read $fh, $buf, unpack("N", $buf);
+    $nfiles++;
 
-    (my $fullname = $buf) =~ s|^([a-f\d]{8})/||;        # strip CRC
-    print $fullname, "\n";
+    $buf = safe_read($fh, 4);
+    $buf = safe_read($fh, unpack("N", $buf));
 
-    read $fh, $buf, 4;
-    read $fh, $buf, unpack("N", $buf);
+    my ($crc, $fullname) = $buf =~ m|^((?i)[a-f\d]{8})/(.*)$|
+        or die qq[unrecognized FILE spec: "$buf"];
+    print "$crc  $fullname\n";
+
+    $buf = safe_read($fh, 4);
+    $buf = safe_read($fh, unpack("N", $buf));
 
     if ($extract)
     {
@@ -73,8 +97,9 @@ while ($buf eq "FILE")
         print STDERR qq[... extracted to $file\n];
     }
 
-    read $fh, $buf, 4;
+    $buf = safe_read($fh, 4);
 }
+printf STDERR qq[$nfiles embedded files found\n];
 
 close $fh;
 
