@@ -154,8 +154,38 @@ followed by a 8-bytes magic string: "C<\012PAR.pm\012>".
 
 =cut
 
+my ($PAR_MAGIC, $FILE_offset_size, $cache_name_size);
+# NOTE: must initialize them in BEGIN as they are used in BEGIN below
+BEGIN {
+    $PAR_MAGIC = "\nPAR.pm\n";
+    $FILE_offset_size = 4;   # pack("N")
+    $cache_name_size = 40;
+    $PKZIP_MAGIC = "PK\003\004";
+}
 
-my ($PAR_MAGIC, $par_temp, $progname, @tmpfile, %ModuleCache);
+
+# Search $fh for the "\nPAR.pm\n signature backward from the end of the file
+sub find_par_magic
+{
+    my ($fh) = @_;
+
+    my $chunk_size = 64 * 1024;
+    my $buf;
+    my $size = -s $fh;
+
+    my $pos = ($size-1) - ($size-1) % $chunk_size;      # NOTE: $pos is a multiple of $chunk_size
+    while ($pos >= 0) {
+        seek $fh, $pos, 0;
+        read $fh, $buf, $chunk_size + length($PAR_MAGIC);
+        if ((my $i = rindex($buf, $PAR_MAGIC)) >= 0) {
+            return $pos + $i;
+        }
+        $pos -= $chunk_size;
+    }
+    return -1;
+}
+
+my ($par_temp, $progname, @tmpfile, %ModuleCache);
 END { if ($ENV{PAR_CLEAN}) {
     require File::Temp;
     require File::Basename;
@@ -204,9 +234,8 @@ rm '$filename'
     }
 } }
 
-BEGIN {
-    $PAR_MAGIC = "\nPAR.pm\n";
 
+BEGIN {
     Internals::PAR::BOOT() if defined &Internals::PAR::BOOT;
 
     eval {
@@ -229,7 +258,7 @@ _set_par_temp();
 outs(qq[\$ENV{PAR_TEMP} = "$ENV{PAR_TEMP}"]);
 
 # Magic string checking and extracting bundled modules {{{
-my ($start_pos, $data_pos);
+my ($start_pos, $start_of_FILE_section);
 MAGIC: {
     local $SIG{__WARN__} = sub {};
 
@@ -240,31 +269,30 @@ MAGIC: {
     }
 
     # Search for the "\nPAR.pm\n signature backward from the end of the file
-    my $chunk_size = 64 * 1024;
-    my $buf;
-    my $size = -s _FH;
-
-    my $magic_pos = $size - $size % $chunk_size; # NOTE: $magic_pos is a multiple of $chunk_size
-    while ($magic_pos >= 0) {
-        seek _FH, $magic_pos, 0;
-        read _FH, $buf, $chunk_size + length($PAR_MAGIC);
-        if ((my $i = rindex($buf, $PAR_MAGIC)) >= 0) {
-            $magic_pos += $i;
-            last;
-        }
-        $magic_pos -= $chunk_size;
-    }
+    my $magic_pos = find_par_magic(*_FH);
     if ($magic_pos < 0) {
         outs(qq[Can't find magic string "$PAR_MAGIC" in file "$progname"]);
         last MAGIC;
     }
+    outs("Found PAR magic at position $magic_pos");
+
+    # Check for "\0CACHE" 4 bytes below the signature
+    seek _FH, $magic_pos - $FILE_offset_size - length("\0CACHE"), 0;
+    read _FH, $buf, length("\0CACHE");
+    if ($buf ne "\0CACHE" && $buf ne "\0CLEAN") {
+        outs("No cache marker found");
+        last MAGIC;
+    }
+    outs(qq[Cache marker "$buf" found]);
 
     # Seek 4 bytes backward from the signature to get the offset of the
     # first embedded FILE, then seek to it
-    seek _FH, $magic_pos - 4, 0;
-    read _FH, $buf, 4;
-    seek _FH, $magic_pos - 4 - unpack("N", $buf), 0;
-    $data_pos = tell _FH;
+    seek _FH, $magic_pos - $FILE_offset_size, 0;
+    read _FH, $buf, $FILE_offset_size;
+    my $offset = unpack("N", $buf);
+    outs("Offset from start of FILEs is $offset");
+    seek _FH, $magic_pos - $FILE_offset_size - $offset, 0;
+    $start_of_FILE_section = tell _FH;
 
     # }}}
 
@@ -288,10 +316,6 @@ MAGIC: {
             $PAR::Heavy::FullCache{$fullname} = $filename;
             $PAR::Heavy::FullCache{$filename} = $fullname;
         }
-        elsif ( $fullname =~ m|^/?shlib/| and defined $ENV{PAR_TEMP} ) {
-            my $filename = _save_as("$basename$ext", $buf, 0755);
-            outs("SHLIB: $filename\n");
-        }
         else {
             $ModuleCache{$fullname} = {
                 buf => $buf,
@@ -364,7 +388,7 @@ MAGIC: {
 
     # }}}
 
-    unless ($buf eq "PK\003\004") {
+    if ($buf ne $PKZIP_MAGIC) {
         outs(qq[No zip found after FILE section in file "$progname"]);
         last MAGIC ;
     }
@@ -379,7 +403,6 @@ my ($out, $bundle, $logfh, $cache_name);
 
 delete $ENV{PAR_APP_REUSE}; # sanitize (REUSE may be a security problem)
 
-$quiet = 0 unless $ENV{PAR_DEBUG};
 # Don't swallow arguments for compiled executables without --par-options
 if (!$start_pos or ($ARGV[0] eq '--par-options' && shift)) {
     my %dist_cmd = qw(
@@ -474,7 +497,8 @@ if ($out) {
     }
 
 
-    my %env = do {
+    # Extract the "par" dictionary from META.yml in $zip
+    my %meta_par = do {
         if ($zip and my $meta = $zip->contents('META.yml')) {
             $meta =~ s/.*^par:$//ms;
             $meta =~ s/^\S.*//ms;
@@ -488,7 +512,7 @@ if ($out) {
         open my $ph, '<:raw', $par or die qq[Can't read par file "$par": $!];
         my $buf;
         read $ph, $buf, 4;
-        die qq["$par" is not a par file] unless $buf eq "PK\003\004";
+        die qq["$par" is not a par file] unless $buf eq $PKZIP_MAGIC;
         close $ph;
     }
 
@@ -496,7 +520,7 @@ if ($out) {
 
     my $fh = IO::File->new(
         $out,
-        IO::File::O_CREAT() | IO::File::O_WRONLY() | IO::File::O_TRUNC(),
+        IO::File::O_CREAT() | IO::File::O_RDWR() | IO::File::O_TRUNC(),
         0777,
     ) or die qq[Can't create file "$out": $!];
     $fh->binmode();
@@ -504,8 +528,8 @@ if ($out) {
     seek _FH, 0, 0;
 
     my $loader;
-    if (defined $data_pos) {
-        read _FH, $loader, $data_pos;
+    if (defined $start_of_FILE_section) {
+        read _FH, $loader, $start_of_FILE_section;
     } else {
         local $/ = undef;
         $loader = <_FH>;
@@ -515,16 +539,9 @@ if ($out) {
         require PAR::Filter::PodStrip;
         PAR::Filter::PodStrip->apply(\$loader, $0);
     }
-    foreach my $key (sort keys %env) {
-        my $val = $env{$key} or next;
-        $val = eval $val if $val =~ /^['"]/;
-        my $magic = "__ENV_PAR_" . uc($key) . "__";
-        my $set = "PAR_" . uc($key) . "=$val";
-        $loader =~ s{$magic( +)}{
-            $magic . $set . (' ' x (length($1) - length($set)))
-        }eg;
-    }
-    $fh->print($loader);
+
+    $fh->print($loader)
+        or die qq[Error writing loader to "$out": $!];
     # }}}
 
     # Write bundled modules {{{
@@ -595,9 +612,9 @@ if ($out) {
             }
             else {
                 local $/ = undef;
-                open my $fh, '<:raw', $file or die qq[Can't read "$file": $!];
-                $content = <$fh>;
-                close $fh;
+                open my $th, '<:raw', $file or die qq[Can't read "$file": $!];
+                $content = <$th>;
+                close $th;
 
                 PAR::Filter::PodStrip->apply(\$content, "<embedded>/$name")
                     if !$ENV{PAR_VERBATIM} and $name =~ /\.(?:pm|ix|al)$/i;
@@ -609,30 +626,44 @@ if ($out) {
                        pack('N', length($name) + 9),
                        sprintf("%08x/%s", Archive::Zip::computeCRC32($content), $name),
                        pack('N', length($content)),
-                       $content);
+                       $content)
+                or die qq[Error writing embedded FILE to "$out": $!];
+
             outs(qq[Written as "$name"]);
         }
     }
     # }}}
 
     # Now write out the PAR and magic strings {{{
-    $zip->writeToFileHandle($fh) if $zip;
-
-    $cache_name = substr $cache_name, 0, 40;
-    if (!$cache_name and my $mtime = (stat($out))[9]) {
-        my $ctx = Digest::SHA->new(1);
-        open my $fh, "<:raw", $out;
-        $ctx->addfile($fh);
-        close $fh;
-
-        $cache_name = $ctx->hexdigest;
+    if ($zip) {
+        $zip->writeToFileHandle($fh) == Archive::Zip::AZ_OK
+            or die qq[Error writing zip part of "$out"];
     }
-    $cache_name .= "\0" x (41 - length $cache_name);
-    $cache_name .= "CACHE";
-    $fh->print($cache_name);
-    $fh->print(pack('N', $fh->tell - length($loader)));
-    $fh->print($PAR_MAGIC);
-    $fh->close;
+
+    if ($meta_par{clean}) {
+        $fh->print("\0CLEAN");
+    }
+    else {
+        if (!defined $cache_name) {
+            my $ctx = Digest::SHA->new(1);
+            seek $fh, 0, 0;
+            $ctx->addfile($fh);
+            seek $fh, 0, 2;
+            $cache_name = $ctx->hexdigest;
+        }
+        # truncate $cache_name to 40 byes and fill with NULs if necessary
+        $cache_name = substr $cache_name, 0, $cache_name_size;
+        $cache_name .= "\0" x ($cache_name_size - length $cache_name);
+        $fh->print($cache_name, 
+                   "\0CACHE");
+    }
+
+    # compute the offset from the end of $loader to end of "...\0CACHE" or "\0CLEAN"
+    my $offset = $fh->tell - length($loader);
+    $fh->print(pack('N', $offset),
+               $PAR_MAGIC);
+
+    $fh->close or die qq[Error writing trailer of "$out": $!];
     chmod 0755, $out;
     # }}}
 
@@ -664,15 +695,15 @@ if ($out) {
     # even if lots of stuff has been appended to the pp'ed exe (e.g. by OSX codesign).
     Archive::Zip::setChunkSize(-s _FH);
     my $zip = Archive::Zip->new;
-    ( $zip->readFromFileHandle($fh, $progname) == Archive::Zip::AZ_OK() )
+    ($zip->readFromFileHandle($fh, $progname) == Archive::Zip::AZ_OK())
         or die qq[Error reading zip archive "$progname"];
     Archive::Zip::setChunkSize(64 * 1024);
 
     push @PAR::LibCache, $zip;
     $PAR::LibCache{$progname} = $zip;
 
-    $quiet = !$ENV{PAR_DEBUG};
 
+    outs("Extracting zip...");
     if (defined $ENV{PAR_TEMP}) { # should be set at this point!
         foreach my $member ( $zip->members ) {
             next if $member->isDirectory;
@@ -680,7 +711,6 @@ if ($out) {
             next unless $member_name =~ m{
                 ^
                 /?shlib/
-                (?:$Config::Config{version}/)?
                 (?:$Config::Config{archname}/)?
                 ([^/]+)
                 $
@@ -690,8 +720,9 @@ if ($out) {
             if (-f $dest_name && -s _ == $member->uncompressedSize()) {
                 outs(qq[Skipping "$member_name" since it already exists at "$dest_name"]);
             } else {
-                outs(qq[Extracting "$member_name" to "$dest_name"]);
-                $member->extractToFileNamed($dest_name);
+                outs(qq[Extracting "$member_name" to "$dest_name"...]);
+                ($member->extractToFileNamed($dest_name) == Archive::Zip::AZ_OK())
+                    or die qq[Error extracting zip member to "$dest_name"];
                 chmod(0555, $dest_name) if $^O eq "hpux";
             }
         }
@@ -786,43 +817,42 @@ sub _set_par_temp {
         else {
             $username = $ENV{USERNAME} || $ENV{USER} || 'SYSTEM';
         }
-        $username =~ s/\W/_/g;
 
         my $stmpdir = "$path$sys{_delim}par-".unpack("H*", $username);
         mkdir $stmpdir, 0755;
-        if (!$ENV{PAR_CLEAN} and my $mtime = (stat($progname))[9]) {
-            open my $fh, "<:raw", $progname or die qq[Can't read "$progname": $!];
-            seek $fh, -18, 2;
-            my $buf;
-            read $fh, $buf, 6;
-            if ($buf eq "\0CACHE") {
-                seek $fh, -58, 2;
-                read $fh, $buf, 41;
-                $buf =~ s/\0//g;
-                $stmpdir .= "$sys{_delim}cache-$buf";
-            }
-            else {
-                my $digest = eval
-                {
-                    require Digest::SHA;
-                    my $ctx = Digest::SHA->new(1);
-                    open my $fh, "<:raw", $progname or die qq[Can't read "$progname": $!];
-                    $ctx->addfile($fh);
-                    close($fh);
-                    $ctx->hexdigest;
-                } || $mtime;
 
-                $stmpdir .= "$sys{_delim}cache-$digest";
-            }
-            close($fh);
+        my $cache_dir;
+        if ($ENV{PAR_CLEAN}) {
+            $cache_dir = "temp-$$";
         }
         else {
-            $ENV{PAR_CLEAN} = 1;
-            $stmpdir .= "$sys{_delim}temp-$$";
+            open my $fh, "<:raw", $progname or die qq[Can't read "$progname": $!];
+            if ((my $magic_pos = find_par_magic($fh)) >= 0) {
+                seek $fh, $magic_pos 
+                          - $FILE_offset_size 
+                          - length("\0CACHE"), 0;
+                my $buf;
+                read $fh, $buf, length("\0CACHE");
+                if ($buf eq "\0CACHE") {
+                    seek $fh, $magic_pos 
+                              - $FILE_offset_size 
+                              - length("\0CACHE") 
+                              - $cache_name_size, 0;
+                    read $fh, $buf, $cache_name_size;
+                    $buf =~ s/\0//g;
+                    $cache_dir = "cache-$buf";
+                }
+            }
+            close $fh;
         }
+        if (!$cache_dir) {
+            $cache_dir = "temp-$$";
+            $ENV{PAR_CLEAN} = 1;
+        }
+        $stmpdir .= "$sys{_delim}$cache_dir";
 
-        $ENV{PAR_TEMP} = $stmpdir;
         mkdir $stmpdir, 0755;
+        $ENV{PAR_TEMP} = $stmpdir;
         last;
     }
 
@@ -842,9 +872,11 @@ sub _save_as {
     unless (-e $fullname) {
         my $tempname = "$fullname.$$";
 
-        open my $fh, '>:raw', $tempname or die qq[Can't write "$tempname": $!];
-        print $fh $contents;
-        close $fh;
+        my $fh;
+        (open $fh, '>:raw', $tempname)
+        && (print $fh $contents)
+        && (close $fh)
+            or die qq[Error writing "$tempname": $!];
         chmod $mode, $tempname if defined $mode;
 
         rename($tempname, $fullname) or unlink($tempname);
@@ -912,21 +944,16 @@ sub _par_init_env {
     }
 
     for (qw( SPAWNED TEMP CLEAN DEBUG CACHE PROGNAME ) ) {
-        delete $ENV{'PAR_'.$_};
-    }
-    for (qw/ TMPDIR TEMP CLEAN DEBUG /) {
-        $ENV{'PAR_'.$_} = $ENV{'PAR_GLOBAL_'.$_} if exists $ENV{'PAR_GLOBAL_'.$_};
+        delete $ENV{"PAR_$_"};
     }
 
-    my $par_clean = "__ENV_PAR_CLEAN__               ";
+    for (qw/ TMPDIR TEMP CLEAN DEBUG /) {
+        $ENV{"PAR_$_"} = $ENV{"PAR_GLOBAL_$_"} if exists $ENV{"PAR_GLOBAL_$_"};
+    }
 
     if ($ENV{PAR_TEMP}) {
         delete $ENV{PAR_CLEAN};
     }
-    elsif (!exists $ENV{PAR_GLOBAL_CLEAN}) {
-        my $value = substr($par_clean, 12 + length("CLEAN"));
-        $ENV{PAR_CLEAN} = $1 if $value =~ /^PAR_CLEAN=(\S+)/;
-    }
 }
 
 sub outs {
@@ -986,8 +1013,10 @@ Steffen Mueller E<lt>smueller@cpan.orgE<gt>
 You can write
 to the mailing list at E<lt>par@perl.orgE<gt>, or send an empty mail to
 E<lt>par-subscribe@perl.orgE<gt> to participate in the discussion.
+Archives of the mailing list are available at
+E<lt>https://www.mail-archive.com/par@perl.org/E<gt> or E<lt>https://groups.google.com/g/perl.parE<gt>.
 
-Please submit bug reports to E<lt>bug-par-packer@rt.cpan.orgE<gt>.
+Please submit bug reports to E<lt>https://github.com/rschupp/PAR-Packer/issuesE<gt>.
 
 =head1 COPYRIGHT
 
